/*
Copyright (c) 2013, Sean Kasun
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

* Redistributions of source code must retain the above copyright notice, this
  list of conditions and the following disclaimer.

* Redistributions in binary form must reproduce the above copyright notice,
  this list of conditions and the following disclaimer in the documentation
  and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
THE POSSIBILITY OF SUCH DAMAGE.
*/

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include "lzx.h"

static inline uint16_t r16(const uint8_t *p)
{
	uint16_t r=p[0]|(p[1]<<8);
	return r;
}
static inline uint32_t r32(const uint8_t *p)
{
	uint32_t r=p[0]|(p[1]<<8)|(p[2]<<16)|(p[3]<<24);
	return r;
}

typedef enum
{
	SCALAR,
	VECTOR,
	MATRIX_ROWS,
	MATRIX_COLUMNS,
	OBJECT,
	STRUCT
} ParameterClass;
static const char *classes[]={
	"scalar",
	"vector",
	"matrix rows",
	"matrix columns",
	"object",
	"struct"
};
typedef enum
{
	VOID,
	BOOL,
	INT,
	FLOAT,
	STRING,
	TEXTURE,
	TEXTURE1D,
	TEXTURE2D,
	TEXTURE3D,
	TEXTURECUBE,
	SAMPLER,
	SAMPLER1D,
	SAMPLER2D,
	SAMPLER3D,
	SAMPLERCUBE,
	PIXELSHADER,
	VERTEXSHADER,
	PIXELFRAGMENT,
	VERTEXFRAGMENT,
	UNSUPPORTED
} ParameterType;
static const char *types[]={
	"void",
	"bool",
	"int",
	"float",
	"string",
	"texture",
	"texture1d",
	"texture2d",
	"texture3d",
	"texturecube",
	"sampler",
	"sampler1d",
	"sampler2d",
	"sampler3d",
	"samplercube",
	"pixelshader",
	"vertexshader",
	"pixelfragment",
	"vertexfragment",
	"unsupported"
};
typedef enum
{
	CONSTANT,
	PARAMETER,
	FXLC
} StateType;

#define ANNOTATION 4

#define OP_MASK 0xffff
typedef enum
{
	OP_NOP			= 0x00,
	OP_MOV			= 0x01,
	OP_ADD			= 0x02,
	OP_SUB			= 0x03,
	OP_MAD			= 0x04,
	OP_MUL			= 0x05,
	OP_RCP			= 0x06,
	OP_RSQ			= 0x07,
	OP_DP3			= 0x08,
	OP_DP4			= 0x09,
	OP_MIN			= 0x0a,
	OP_MAX			= 0x0b,
	OP_SLT			= 0x0c,
	OP_SGE			= 0x0d,
	OP_EXP			= 0x0e,
	OP_LOG			= 0x0f,
	OP_LIT			= 0x10,
	OP_DST			= 0x11,
	OP_LRP			= 0x12,
	OP_FRC			= 0x13,
	OP_M4x4			= 0x14,
	OP_M4x3			= 0x15,
	OP_M3x4			= 0x16,
	OP_M3x3			= 0x17,
	OP_M3x2			= 0x18,
	OP_CALL			= 0x19,
	OP_CALLNZ		= 0x1a,
	OP_LOOP			= 0x1b,
	OP_RET			= 0x1c,
	OP_ENDLOOP		= 0x1d,
	OP_LABEL		= 0x1e,
	OP_DCL			= 0x1f,
	OP_POW			= 0x20,
	OP_CRS			= 0x21,
	OP_SGN			= 0x22,
	OP_ABS			= 0x23,
	OP_NRM			= 0x24,
	OP_SINCOS		= 0x25,
	OP_REP			= 0x26,
	OP_ENDREP		= 0x27,
	OP_IF			= 0x28,
	OP_IFC			= 0x29,
	OP_ELSE			= 0x2a,
	OP_ENDIF		= 0x2b,
	OP_BREAK		= 0x2c,
	OP_BREAKC		= 0x2d,
	OP_MOVA			= 0x2e,
	OP_DEFB			= 0x2f,
	OP_DEFI			= 0x30,
	OP_TEXCOORD		= 0x40,
	OP_TEXKILL		= 0x41,
	OP_TEX			= 0x42,
	OP_TEXBEM		= 0x43,
	OP_TEXBEML		= 0x44,
	OP_TEXREG2AR	= 0x45,
	OP_TEXREG2GB	= 0x46,
	OP_TEXM3x2PAD	= 0x47,
	OP_TEXM3x2TEX	= 0x48,
	OP_TEXM3x3PAD	= 0x49,
	OP_TEXM3x3TEX	= 0x4a,
	OP_TEXM3x3DIFF	= 0x4b,
	OP_TEXM3x3SPEC	= 0x4c,
	OP_TEXM3x3VSPEC	= 0x4d,
	OP_EXPP			= 0x4e,
	OP_LOGP			= 0x4f,
	OP_CND			= 0x50,
	OP_DEF			= 0x51,
	OP_TEXREG2RGB	= 0x52,
	OP_TEXDP3TEX	= 0x53,
	OP_TEXM3x2DEPTH	= 0x54,
	OP_TEXDP3		= 0x55,
	OP_TEXM3x3		= 0x56,
	OP_TEXDEPTH		= 0x57,
	OP_CMP			= 0x58,
	OP_BEM			= 0x59,
	OP_DP2ADD		= 0x5a,
	OP_DSX			= 0x5b,
	OP_DSY			= 0x5c,
	OP_TEXLDD		= 0x5d,
	OP_SETP			= 0x5e,
	OP_TEXLDL		= 0x5f,
	OP_BREAKP		= 0x60,
	OP_PHASE		= 0xfffd,
	OP_COMMENT		= 0xfffe,
	OP_END			= 0xffff,
	OP_ENDOFTABLE	= 0x10000
} Opcode;

const char *opcodeNames[]={
	"nop", //00
	"mov", //01
	"add", //02
	"sub", //03
	"mad", //04
	"mul", //05
	"rcp", //06
	"rsq", //07
	"dp3", //08
	"dp4", //09
	"min", //0a
	"max", //0b
	"slt", //0c
	"sge", //0d
	"exp", //0e
	"log", //0f
	"lit", //10
	"dst", //11
	"lrp", //12
	"frc", //13
	"m4x4", //14
	"m4x3", //15
	"m3x4", //16
	"m3x3", //17
	"m3x2", //18
	"call", //19
	"callnz", //1a
	"loop", //1b
	"ret", //1c
	"endloop", //1d
	"label", //1e
	"dcl", //1f
	"pow", //20
	"crs", //21
	"sgn", //22
	"abs", //23
	"nrm", //24
	"sincos", //25
	"rep", //26
	"endrep", //27
	"if", //28
	"ifc", //29
	"else", //2a
	"endif", //2b
	"break", //2c
	"breakc", //2d
	"mova", //2e
	"defb", //2f
	"defi", //30
	"", //31
	"", //32
	"", //33
	"", //34
	"", //35
	"", //36
	"", //37
	"", //38
	"", //39
	"", //3a
	"", //3b
	"", //3c
	"", //3d
	"", //3e
	"", //3f
	"texcrd", //40
	"texkill", //41
	"texld", //42
	"texbem", //43
	"texbeml", //44
	"texreg2ar", //45
	"texreg2gb", //46
	"texm3x2pad", //47
	"texm3x2tex", //48
	"texm3x3pad", //49
	"texm3x3tex", //4a
	"texm3x3diff", //4b
	"texm3x3spec", //4c
	"texm3x3vspec", //4d
	"expp", //4e
	"logp", //4f
	"cnd", //50
	"def", //51
	"texreg2rgb", //52
	"texdp3tex", //53
	"texm3x2depth", //54
	"texdp3", //55
	"texm3x3", //56
	"texdepth", //57
	"cmp", //58
	"bem", //59
	"dp2add", //5a
	"dsx", //5b
	"dsy", //5c
	"texldd", //5d
	"setp", //5e
	"texldl", //5f
	"breakp" //60
};

static struct {
	Opcode opcode;
	int numDest;
	int numSrc;
} opcodes[]={
	{OP_NOP,		0,0},
	{OP_MOV,		1,1},
	{OP_ADD,		1,2},
	{OP_SUB,		1,2},
	{OP_MAD,		1,3},
	{OP_MUL,		1,2},
	{OP_RCP,		1,1},
	{OP_RSQ,		1,1},
	{OP_DP3,		1,2},
	{OP_DP4,		1,2},
	{OP_MIN,		1,2},
	{OP_MAX,		1,2},
	{OP_SLT,		1,2},
	{OP_SGE,		1,2},
	{OP_ABS,		1,1},
	{OP_EXP,		1,1},
	{OP_LOG,		1,1},
	{OP_EXPP,		1,1},
	{OP_LOGP,		1,1},
	{OP_DST,		1,2},
	{OP_LRP,		1,3},
	{OP_FRC,		1,1},
	{OP_CND,		1,3},
	{OP_CMP,		1,3},
	{OP_POW,		1,2},
	{OP_CRS,		1,2},
	{OP_NRM,		1,1},
	{OP_SINCOS,		1,3},
	{OP_DP2ADD,		1,3},
	{OP_M4x4,		1,2},
	{OP_M4x3,		1,2},
	{OP_M3x4,		1,2},
	{OP_M3x3,		1,2},
	{OP_M3x2,		1,2},
	{OP_DCL,		0,2},
	{OP_REP,		0,1},
	{OP_ENDREP,		0,0},
	{OP_IF,			0,1},
	{OP_IFC,		0,2},
	{OP_ELSE,		0,0},
	{OP_ENDIF,		0,0},
	{OP_BREAK,		0,0},
	{OP_BREAKC,		0,2},
	{OP_BREAKP,		0,1},
	{OP_CALL,		0,1},
	{OP_CALLNZ,		0,2},
	{OP_LOOP,		0,2},
	{OP_RET,		0,0},
	{OP_ENDLOOP,	0,0},
	{OP_LABEL,		0,1},
	{OP_DEF,		1,4},
	{OP_DEFB,		1,1},
	{OP_DEFI,		1,4},
	{OP_TEXCOORD,	1,1},
	{OP_TEXKILL,	1,0},
	{OP_TEX,		1,2},
	{OP_TEXBEM,		1,1},
	{OP_TEXBEML,	1,1},
	{OP_TEXREG2AR,	1,1},
	{OP_TEXREG2GB,	1,1},
	{OP_TEXREG2RGB,	1,1},
	{OP_TEXM3x2PAD,	1,1},
	{OP_TEXM3x2TEX,	1,1},
	{OP_TEXM3x3PAD,	1,1},
	{OP_TEXM3x3DIFF,1,1},
	{OP_TEXM3x3SPEC,1,2},
	{OP_TEXM3x3VSPEC,1,1},
	{OP_TEXM3x3TEX,	1,1},
	{OP_TEXDP3TEX,	1,1},
	{OP_TEXM3x2DEPTH,1,1},
	{OP_TEXDP3,		1,1},
	{OP_TEXM3x3,	1,1},
	{OP_TEXDEPTH,	1,0},
	{OP_BEM,		1,2},
	{OP_DSX,		1,1},
	{OP_DSY,		1,1},
	{OP_TEXLDD,		1,4},
	{OP_SETP,		1,2},
	{OP_TEXLDL,		1,2},
	{OP_PHASE,		0,0},
	{OP_ENDOFTABLE,	0,0}
};

typedef enum {
	REG_TEMP = 0,
	REG_INPUT = 1,
	REG_CONST = 2,
	REG_ADDR = 3,
	REG_TEXTURE = 3,
	REG_RASTOUT = 4,
	REG_ATTROUT = 5,
	REG_TEXCRDOUT = 6,
	REG_OUTPUT = 6,
	REG_CONSTINT = 7,
	REG_COLOROUT = 8,
	REG_DEPTHOUT = 9,
	REG_SAMPLER = 10,
	REG_CONST2 = 11,
	REG_CONST3 = 12,
	REG_CONST4 = 13,
	REG_CONSTBOOL = 14,
	REG_LOOP = 15,
	REG_TEMPFLOAT16 = 16,
	REG_MISCTYPE = 17,
	REG_LABEL = 18,
	REG_PREDICATE = 19,
	REG_IMMCONST,
	REG_CONSTBUFFER,
	REG_PRIMID,
	REG_NULL,
	REG_RESOURCE
} RegisterType;

typedef enum {
	DATA_FLOAT,
	DATA_INT,
	DATA_RESOURCE,
	DATA_SAMPLER,
	DATA_UINT
} DataType;

typedef enum {
	IMM_SCALAR,
	IMM_VEC4
} ImmediateType;

typedef struct
{
	RegisterType type;
	DataType dataType;
	uint32_t offset;
	ImmediateType immType;
	uint32_t imm[4];
	uint32_t writeMask;
	uint32_t modifiers;
	uint32_t shift;
	uint32_t swizzle;
} Register;

typedef struct _Parameter
{
	char *name,*semantic;
	ParameterType type;
	ParameterClass klass;
	uint32_t flags;
	uint32_t size;
	uint32_t numAnnotations,numMembers,numElements;
	uint32_t rows,columns;
	void *data;
	struct _Parameter *children;
	struct _Parameter *annotations;
} Parameter; 
typedef struct
{
	uint32_t operation;
	uint32_t index;
	StateType type;
	Parameter *param;
} State;

typedef struct
{
	uint32_t numStates;
	State *states;
} Sampler;

typedef struct
{
	char *name;
	uint32_t numAnnotations,numStates;
	Parameter *annotations;
	State *states;
} Pass;

typedef struct
{
	char *name;
	uint32_t numAnnotations,numPasses;
	Parameter *annotations;
	Pass *passes;
} Technique;

static uint8_t *parseValue(Parameter *param,uint8_t *ptr,uint8_t *constData,Parameter **objects,int padlen);

static void pad(int len)
{
	int i;
	for (i=0;i<len;i++)
		printf("    ");
}

static char *parseName(uint8_t *ptr)
{
	uint32_t size=r32(ptr);
	char *name=NULL;
	if (size==0)
	{
		name=malloc(1);
		name[0]=0;
	}
	else
	{
		name=malloc(size);
		memcpy(name,ptr+4,size);
	}
	return name;
}

static void parseVertexShader(uint8_t *ptr,void *data,int padlen)
{
	fprintf(stderr,"parseVertexShader\n");
}
static const char *commentStr(const char *str, int n)
{
	static const char hex[16]="0123456789abcdef";
	static char buffer[310];

	size_t size=10+n*4;
	if (size>310) size=310;
	char *dst=buffer;
	*dst++='"';
	while (n-->0 && dst<=buffer+size-9)
	{
		char c=*str++;
		switch (c)
		{
		case '\n': *dst++='\\'; *dst++='n'; break;
		case '\r': *dst++='\\'; *dst++='r'; break;
		case '\t': *dst++='\\'; *dst++='t'; break;
		case '"': *dst++='\\'; *dst++='"'; break;
		case '\\': *dst++='\\'; *dst++='\\'; break;
		default:
				   if (c>=' ' && c<=126)
					   *dst++=c;
				   else
				   {
					   *dst++='\\';
					   *dst++='x';
					   *dst++=hex[(c>>4)&0xf];
					   *dst++=hex[c&0xf];
				   }
		}
	}
	*dst++='"';
	if (n>0)
	{
		*dst++='.';
		*dst++='.';
		*dst++='.';
	}
	*dst++=0;
	return buffer;
}
static uint8_t *readShaderComment(uint8_t *ptr,int padlen)
{
	uint32_t token=r32(ptr);
	while ((token&OP_MASK)==OP_COMMENT)
	{
		uint32_t size=(token&0x7fff0000)>>16;
		const char *comment=ptr+4;
		ptr+=size*4+4;
		if (size>1 && r32(comment)==0x47554244) //DBUG
		{
			uint32_t ofs=r32(comment+0x28);
			pad(padlen);
			printf("// %s\n",comment+4+ofs);
		}
		else if (size)
		{
			pad(padlen);
			printf("// %s\n",commentStr(comment,size*4));
		}
		else
			break;
		token=r32(ptr);
	}
	return ptr;
}
static uint8_t *shaderEnd(uint8_t *ptr,int padlen)
{
	ptr=readShaderComment(ptr,padlen);
	if (r32(ptr)==OP_END)
		return NULL;
	return ptr;
}
static uint8_t *readDest(uint8_t *ptr,Register *reg)
{
	uint32_t token=r32(ptr); ptr+=4;
	reg->type=((token&0x70000000)>>28)|((token&0x1800)>>8);
	reg->dataType=DATA_FLOAT;
	reg->offset=token&0x7ff;
	reg->writeMask=(token&0xf0000)>>16;
	reg->modifiers=(token&0xf00000)>>20;
	reg->shift=(token&0xf000000)>>24;
	return ptr;
}
static uint8_t *readSrc(uint8_t *ptr,Register *reg)
{
	uint32_t token=r32(ptr); ptr+=4;
	reg->type=((token&0x70000000)>>28)|((token&0x1800)>>8);
	reg->dataType=DATA_FLOAT;
	reg->offset=token&0x7ff;
	reg->swizzle=(token&0xff0000)>>16;
	reg->modifiers=(token&0xf000000)>>24;
	return ptr;
}
static uint8_t *readImmediate(uint8_t *ptr,Register *imm,ImmediateType immtype,DataType datatype)
{
	imm->type=REG_IMMCONST;
	imm->dataType=datatype;
	imm->offset=0;
	imm->immType=immtype;
	int i;
	for (i=0;i<(immtype==IMM_VEC4?4:1);i++)
	{
		imm->imm[i]=r32(ptr); ptr+=4;
	}
	imm->swizzle=0xe4; //no swizzle
	imm->modifiers=0;
	return ptr;
}
static void printReg(Register *reg)
{
	static const char * const rastout_regs[]={"oPos","oFog","oPts"};
	static const char * const misc_regs[]={"vPos","vFace"};
	uint32_t offset=reg->offset;
	switch (reg->type)
	{
	case REG_TEMP: printf("r"); break;
	case REG_INPUT: printf("v"); break;
	case REG_CONST: printf("c"); break;
	case REG_CONST2: printf("c"); offset+=2048; break;
	case REG_CONST3: printf("c"); offset+=4096; break;
	case REG_CONST4: printf("c"); offset+=6144; break;
	case REG_TEXTURE: printf("t"); break;
	case REG_RASTOUT: printf("%s",rastout_regs[offset]); break;
	case REG_COLOROUT: printf("oC"); break;
	case REG_DEPTHOUT: printf("oDepth"); break;
	case REG_ATTROUT: printf("oD"); break;
	case REG_TEXCRDOUT: printf("oT"); break;
	case REG_CONSTINT: printf("i"); break;
	case REG_CONSTBOOL: printf("b"); break;
	case REG_LABEL: printf("l"); break;
	case REG_LOOP: printf("aL"); break;
	case REG_SAMPLER: printf("s"); break;
	case REG_MISCTYPE: printf("%s",misc_regs[offset]); break;
	case REG_PREDICATE: printf("p"); break;
	case REG_IMMCONST: printf("l"); break;
	case REG_CONSTBUFFER: printf("cb"); break;
	case REG_PRIMID: printf("primID"); break;
	case REG_NULL: printf("null"); break;
	case REG_RESOURCE: printf("t"); break;
	default: fprintf(stderr,"unknown type: %d\n",reg->type); exit(1); break;
	}
	if (reg->type==REG_IMMCONST)
	{
		printf("(");
		switch (reg->immType)
		{
		case IMM_SCALAR:
			switch (reg->dataType)
			{
			case DATA_FLOAT:
				{
					union {
						uint32_t d;
						float f;
					} fd;
					fd.d=reg->imm[0];
					printf("%.8e",fd.f);
				}
				break;
			case DATA_INT:
				printf("%d",reg->imm[0]);
				break;
			case DATA_RESOURCE:
			case DATA_SAMPLER:
			case DATA_UINT:
				printf("%u",reg->imm[0]);
				break;
			default:
				fprintf(stderr,"Unknown data type: %d\n",reg->dataType);
				exit(1);
			}
			break;
		case IMM_VEC4:
			switch (reg->dataType)
			{
			case DATA_FLOAT:
				{
					union {
						uint32_t d;
						float f;
					} fd;
					int j;
					for (j=0;j<4;j++)
					{
						fd.d=reg->imm[j];
						printf("%.8e%s",fd.f,j==3?"":", ");
					}
				}
				break;
			case DATA_INT:
				printf("%d, %d, %d, %d",reg->imm[0],reg->imm[1],reg->imm[2],reg->imm[3]);
				break;
			case DATA_RESOURCE:
			case DATA_SAMPLER:
			case DATA_UINT:
				printf("%u, %u, %u, %u",reg->imm[0],reg->imm[1],reg->imm[2],reg->imm[3]);
				break;
			default:
				fprintf(stderr,"Unknown data type: %d\n",reg->dataType);
				exit(1);
			}
			break;
		default:
			fprintf(stderr,"Unknown imm type: %d\n",reg->immType);
			exit(1);
		}
		printf(")");
	}
	else if (reg->type!=REG_RASTOUT && reg->type!=REG_MISCTYPE && reg->type!=REG_NULL)
	{
		if (offset!=0)
			printf("%u",offset);
	}
}
static void printDest(Register *reg)
{
	printReg(reg);
	if (reg->writeMask!=0x0 && reg->writeMask!=0xf)
	{
		printf(".");
		if (reg->writeMask&1) printf("x");
		if (reg->writeMask&2) printf("y");
		if (reg->writeMask&4) printf("z");
		if (reg->writeMask&8) printf("w");
	}
}
static void printSrc(Register *reg)
{
	if (reg->modifiers==1 || reg->modifiers==3 ||
			reg->modifiers==5 || reg->modifiers==8 ||
			reg->modifiers==12)
		printf("-");
	else if (reg->modifiers==6)
		printf("1-");
	else if (reg->modifiers==13)
		printf("!");
	if (reg->modifiers==11 || reg->modifiers==12)
		printf("abs(");
	printReg(reg);
	if (reg->modifiers)
	{
		switch (reg->modifiers)
		{
		case 0: break; //none
		case 1: break; //neg
		case 13: break; //not
		case 2: printf("_bias"); break; //bias
		case 3: printf("_bias"); break; //biasneg
		case 4: printf("_bx2"); break; //sign
		case 5: printf("_bx2"); break; //signneg
		case 6: break; //comp
		case 7: printf("_x2"); break; //x2
		case 8: printf("_x2"); break; //x2neg
		case 9: printf("_dz"); break; //dz
		case 10: printf("_dw"); break; //dw
		case 12: printf(")"); break; //absneg
		case 11: printf(")"); break; //abs
		default: printf("_unknown_modifier(%#x)",reg->modifiers); break;
		}
	}
	if (reg->swizzle!=0xe4) //no swizzle
	{
		static const char *swiz="xyzw";
		uint32_t x=reg->swizzle&3;
		uint32_t y=(reg->swizzle>>2)&3;
		uint32_t z=(reg->swizzle>>4)&3;
		uint32_t w=(reg->swizzle>>6)&3;
		if (x==y && x==z && x==w)
			printf(".%c",swiz[x]);
		else
			printf(".%c%c%c%c",swiz[x],swiz[y],swiz[z],swiz[w]);
	}
}
static void printMods(Register *reg)
{
	switch (reg->shift)
	{
	case 0: break;
	case 13: printf("_d8"); break;
	case 14: printf("_d4"); break;
	case 15: printf("_d2"); break;
	case 1: printf("_x2"); break;
	case 2: printf("_x4"); break;
	case 3: printf("_x8"); break;
	default:
			fprintf(stderr,"Unknown shift: %d\n",reg->shift);
			exit(1);
	}
	if (reg->modifiers&1) printf("_sat");
	if (reg->modifiers&2) printf("_pp");
	if (reg->modifiers&4) printf("_centroid");
}
static void parsePixelShader(uint8_t *ptr,void *data,int padlen)
{
	uint32_t version=r32(ptr);
	if ((version>>16)!=0xffff) //SM1 pixel shader
	{
		fprintf(stderr,"Unknown shader\n");
		return;
	}
	uint8_t major=(version>>8)&0xff;
	uint8_t minor=(version&0xff);
	if (major>3 && minor>0)
	{
		fprintf(stderr,"Invalid shader version %d.%d\n",major,minor);
		return;
	}
	ptr+=4; //skip version
	while ((ptr=shaderEnd(ptr,padlen)))
	{
		uint32_t token=r32(ptr); ptr+=4;

		uint32_t i=0;
		while (opcodes[i].opcode!=OP_ENDOFTABLE)
		{
			if ((token&OP_MASK)==opcodes[i].opcode)
				break;
			i++;
		}
		if (opcodes[i].opcode==OP_ENDOFTABLE)
		{
			fprintf(stderr,"Unknown opcode: %x\n",token);
			exit(1);
		}
		switch (opcodes[i].opcode)
		{
		case OP_DCL:
			{
			uint32_t usageToken=r32(ptr); ptr+=4;
			uint8_t samplerType=(usageToken&0x78000000)>>27;
			Register sem;
			ptr=readDest(ptr,&sem);

			pad(padlen);
			printf("dcl");
			if (sem.type==REG_SAMPLER)
			{
				switch (samplerType)
				{
				case 2: printf("_2d"); break;
				case 3: printf("_cube"); break;
				case 4: printf("_volume"); break;
				default:
						fprintf(stderr,"Unknown type %d\n",samplerType);
						exit(1);
				}
			}
			printMods(&sem);
			printf(" ");
			printDest(&sem);
			printf("\n");
			}
			break;
		case OP_DEF:
			{
				Register dest,imm;
				ptr=readDest(ptr,&dest);
				ptr=readImmediate(ptr,&imm,IMM_VEC4,DATA_FLOAT);

				uint32_t offset=dest.offset;
				switch (dest.type)
				{
				case REG_CONST: break;
				case REG_CONST2: offset+=2048; break;
				case REG_CONST3: offset+=4096; break;
				case REG_CONST4: offset+=6144; break;
				default:
					 fprintf(stderr,"Unhandled register type: %d\n",dest.type);
					 exit(1);
				}
				pad(padlen);
				printf("def c%d =",offset);
				union {
					float f;
					uint32_t l;
				} fl;
				int j;
				for (j=0;j<4;j++)
				{
					fl.l=imm.imm[j];
					printf("%s %f",j==0?"":",",fl.f);
				}
				printf("\n");
			}
			break;
		case OP_DEFI:
			fprintf(stderr,"Unhandled DEFI\n");
			exit(1);
			break;
		case OP_DEFB:
			fprintf(stderr,"Unhandled DEFB\n");
			exit(1);
			break;
		default:
			{
				Register dest,pred,src[4];
				if (opcodes[i].numDest)
					ptr=readDest(ptr,&dest);
				if (token&0x10000000)
					ptr=readSrc(ptr,&pred);
				int j;
				for (j=0;j<opcodes[i].numSrc;j++)
					ptr=readSrc(ptr,&src[j]);

				pad(padlen);
				if (token&0x10000000)
				{
					printf("(");
					printSrc(&pred);
					printf(")");
				}
				if (token&0x40000000)
					printf("+");
				if (opcodes[i].opcode<0x61)
					printf("%s",opcodeNames[opcodes[i].opcode]);
				else if (opcodes[i].opcode==OP_PHASE)
					printf("phase");
				else
				{
					fprintf(stderr,"Unknown opcode: %x\n",opcodes[i].opcode);
					exit(1);
				}

				uint8_t flags=(token&0xff0000)>>16;
				if (opcodes[i].opcode==OP_IFC ||
					opcodes[i].opcode==OP_BREAKC)
				{
					switch (flags)
					{
					case 1: printf("_gt"); break;
					case 2: printf("_eq"); break;
					case 3: printf("_ge"); break;
					case 4: printf("_lt"); break;
					case 5: printf("_ne"); break;
					case 6: printf("_le"); break;
					default: printf("_(%u)",flags); break;
					}
				}
				else if (opcodes[i].opcode==OP_TEX && (flags&1))
					printf("p");
				if (opcodes[i].numDest)
				{
					printMods(&dest);
					printf(" ");
					printDest(&dest);
				}
				for (j=0;j<opcodes[i].numSrc;j++)
				{
					if (j>0 || opcodes[i].numDest)
						printf(",");
					printf(" ");
					printSrc(&src[j]);
				}
				printf("\n");
			}
			break;
		}
	}
}

static uint8_t *parseType(Parameter *param,uint8_t *ptr,uint8_t *constData,Parameter *parent,uint32_t flags,int padlen)
{
	param->flags=flags;
	if (parent==NULL)
	{
		param->type=r32(ptr); ptr+=4;
		param->klass=r32(ptr); ptr+=4;
		uint32_t offset=r32(ptr); ptr+=4;
		param->name=parseName(constData+offset);
		offset=r32(ptr); ptr+=4;
		param->semantic=parseName(constData+offset);
		param->numElements=r32(ptr); ptr+=4;
		param->numMembers=0;
		switch (param->klass)
		{
		case VECTOR:
			param->columns=r32(ptr); ptr+=4;
			param->rows=r32(ptr); ptr+=4;
			param->size=sizeof(uint32_t)*param->rows*param->columns;
			break;
		case SCALAR:
		case MATRIX_ROWS:
		case MATRIX_COLUMNS:
			param->rows=r32(ptr); ptr+=4;
			param->columns=r32(ptr); ptr+=4;
			param->size=sizeof(uint32_t)*param->rows*param->columns;
			break;
		case STRUCT:
			param->numMembers=r32(ptr); ptr+=4;
			break;
		case OBJECT:
			switch (param->type)
			{
			case STRING:
			case PIXELSHADER:
			case VERTEXSHADER:
			case TEXTURE:
			case TEXTURE1D:
			case TEXTURE2D:
			case TEXTURE3D:
			case TEXTURECUBE:
				param->size=sizeof(void *);
				break;
			case SAMPLER:
			case SAMPLER1D:
			case SAMPLER2D:
			case SAMPLER3D:
			case SAMPLERCUBE:
				param->size=0;
				break;
			default:
				fprintf(stderr,"Unknown type: %d\n",param->type);
				return NULL;
			}
			break;
		default:
			fprintf(stderr,"Unknown class: %d\n",param->klass);
			return NULL;
		}
	}
	else
	{
		param->type=parent->type;
		param->klass=parent->klass;
		param->name=parent->name;
		param->semantic=parent->semantic;
		param->numElements=0;
		param->numAnnotations=0;
		param->numMembers=parent->numMembers;
		param->size=parent->size;
		param->rows=parent->rows;
		param->columns=parent->columns;
	}

	pad(padlen); printf("Class %s, Type %s\n",classes[param->klass],types[param->type]);
	pad(padlen); printf("{\n");


	uint32_t i;
	if (param->numElements)
	{
		uint32_t size=0;
		param->children=malloc(sizeof(Parameter)*param->numElements);
		for (i=0;i<param->numElements;i++)
		{
			if (!(ptr=parseType(&param->children[i],ptr,constData,param,flags,padlen+1)))
			{
				fprintf(stderr,"Couldn't parse member #%d\n",i);
				return NULL;
			}
			size+=param->children[i].size;
		}
		param->size=size;
	}
	else if (param->numMembers)
	{
		param->children=malloc(sizeof(Parameter)*param->numMembers);
		for (i=0;i<param->numMembers;i++)
		{
			if (!(ptr=parseType(&param->children[i],ptr,constData,NULL,flags,padlen+1)))
			{
				fprintf(stderr,"Couldn't parse member #%d\n",i);
				return NULL;
			}
			param->size+=param->children[i].size;
		}
	}
	pad(padlen); printf("}\n");
	return ptr;
}
static uint8_t *parseState(State *state,uint8_t *ptr,uint8_t *constData,Parameter **objects,int padlen)
{
	state->param=malloc(sizeof(Parameter));
	state->type=CONSTANT;
	state->operation=r32(ptr); ptr+=4;
	state->index=r32(ptr); ptr+=4;
	uint32_t offset=r32(ptr); ptr+=4;
	if (!parseType(state->param,constData+offset,constData,NULL,0,padlen))
	{
		fprintf(stderr,"Couldn't parse type\n");
		return NULL;
	}
	offset=r32(ptr); ptr+=4;
	if (!parseValue(state->param,constData+offset,constData,objects,padlen))
	{
		fprintf(stderr,"Couldn't parse value\n");
		return NULL;
	}
	return ptr;
}
static uint8_t *parseSampler(Sampler *sampler,uint8_t *ptr,uint8_t *constData,Parameter **objects,int padlen)
{
	sampler->numStates=r32(ptr); ptr+=4;
	sampler->states=malloc(sizeof(State)*sampler->numStates);
	uint32_t i;
	for (i=0;i<sampler->numStates;i++)
	{
		if (!(ptr=parseState(&sampler->states[i],ptr,constData,objects,padlen)))
		{
			fprintf(stderr,"Couldn't parse state #%d\n",i);
			return NULL;
		}
	}
	return ptr;
}
static uint8_t *parseValue2(Parameter *param,uint8_t *ptr,uint8_t *value,uint8_t *constData,Parameter **objects,int padlen)
{
	uint32_t i;
	if (param->numElements)
	{
		param->data=value;
		for (i=0;i<param->numElements;i++)
		{
			if (!(ptr=parseValue2(&param->children[i],ptr,value,constData,objects,padlen+1)))
			{
				fprintf(stderr,"Couldn't parse value #%d\n",i);
				return NULL;
			}
			if (value) value+=param->children[i].size;
		}
		return ptr;
	}

	switch (param->klass)
	{
	case SCALAR:
	case VECTOR:
	case MATRIX_ROWS:
	case MATRIX_COLUMNS:
		param->data=value;
		break;
	case STRUCT:
		param->data=value;
		for (i=0;i<param->numMembers;i++)
		{
			if (!(ptr=parseValue2(&param->children[i],ptr,value,constData,objects,padlen+1)))
			{
				fprintf(stderr,"Couldn't parse value #%d\n",i);
				return NULL;
			}
			if (value) value+=param->children[i].size;
		}
		break;
	case OBJECT:
		switch (param->type)
		{
		case STRING:
		case TEXTURE:
		case TEXTURE1D:
		case TEXTURE2D:
		case TEXTURE3D:
		case TEXTURECUBE:
		case PIXELSHADER:
		case VERTEXSHADER:
			{
				uint32_t id=r32(ptr); ptr+=4;
				objects[id]=param;
				param->data=value;
				break;
			}
		case SAMPLER:
		case SAMPLER1D:
		case SAMPLER2D:
		case SAMPLER3D:
		case SAMPLERCUBE:
			{
				Sampler *sampler=malloc(sizeof(Sampler));
				if (!(ptr=parseSampler(sampler,ptr,constData,objects,padlen)))
				{
					fprintf(stderr,"Couldn't parse sampler\n");
					return NULL;
				}
				param->data=sampler;
				break;
			}
		default:
			fprintf(stderr,"Unknown type: %d\n",param->type);
			break;
		}
		break;
	default:
		fprintf(stderr,"Unknown class: %d\n",param->klass);
		break;
	}
	return ptr;
}
static uint8_t *parseValue(Parameter *param,uint8_t *ptr,uint8_t *constData,Parameter **objects,int padlen)
{
	void *value=NULL;
	if (param->size)
	{
		value=malloc(param->size);
		switch (param->klass)
		{
		case OBJECT:
			break;
		case SCALAR:
		case VECTOR:
		case MATRIX_ROWS:
		case MATRIX_COLUMNS:
			memcpy(value,ptr,param->size);
			break;
		default:
			fprintf(stderr,"Unknown class: %d\n",param->klass);
			return NULL;
		}
	}
	return parseValue2(param,ptr,value,constData,objects,padlen);
}

static uint8_t *parseAnnotation(Parameter *anno,uint8_t *ptr,uint8_t *constData,Parameter **objects,int padlen)
{
	anno->flags=ANNOTATION;
	uint32_t offset=r32(ptr); ptr+=4;
	if (!parseType(anno,constData+offset,constData,NULL,ANNOTATION,padlen))
	{
		fprintf(stderr,"Failed to parse type\n");
		return NULL;
	}
	offset=r32(ptr); ptr+=4;
	if (!parseValue(anno,constData+offset,constData,objects,padlen))
	{
		fprintf(stderr,"Failed to parse value\n");
		return NULL;
	}
	return ptr;
}

static uint8_t *parseParameter(Parameter *param,uint8_t *ptr,uint8_t *constData,Parameter **objects,int padlen)
{
	uint32_t typeofs=r32(ptr); ptr+=4;
	uint32_t valofs=r32(ptr); ptr+=4;
	param->flags=r32(ptr); ptr+=4;
	param->numAnnotations=r32(ptr); ptr+=4;

	pad(padlen); printf("Parameter\n"); pad(padlen); printf("{\n");

	if (!parseType(param,constData+typeofs,constData,NULL,param->flags,padlen+1))
	{
		fprintf(stderr,"Failed to parse type\n");
		return NULL;
	}
	if (!parseValue(param,constData+valofs,constData,objects,padlen+1))
	{
		fprintf(stderr,"Failed to parse value\n");
		return NULL;
	}

	if (param->numAnnotations)
	{
		uint32_t i;
		param->annotations=malloc(sizeof(Parameter)*param->numAnnotations);
		for (i=0;i<param->numAnnotations;i++)
		{
			if (!(ptr=parseAnnotation(&param->annotations[i],ptr,constData,objects,padlen+1)))
			{
				fprintf(stderr,"Failed to parse annotation #%d\n",i);
				return NULL;
			}
		}
	}
	pad(padlen+1); printf("Name: %s\n",param->name);
	pad(padlen); printf("}\n");

	return ptr;
}

static uint8_t *parsePass(Pass *pass,uint8_t *ptr,uint8_t *constData,Parameter **objects,int padlen)
{
	uint32_t offset=r32(ptr); ptr+=4;
	pass->name=parseName(constData+offset);
	pad(padlen); printf("Pass %s\n",pass->name);
	pad(padlen); printf("{\n");
	pass->numAnnotations=r32(ptr); ptr+=4;
	pass->numStates=r32(ptr); ptr+=4;

	uint32_t i;
	if (pass->numAnnotations)
	{
		pass->annotations=malloc(sizeof(Parameter)*pass->numAnnotations);
		for (i=0;i<pass->numAnnotations;i++)
		{
			if (!(ptr=parseAnnotation(&pass->annotations[i],ptr,constData,objects,padlen+1)))
			{
				fprintf(stderr,"Couldn't parse annotation #%d\n",i);
				return NULL;
			}
		}
	}
	if (pass->numStates)
	{
		pass->states=malloc(sizeof(State)*pass->numStates);
		for (i=0;i<pass->numStates;i++)
		{
			if (!(ptr=parseState(&pass->states[i],ptr,constData,objects,padlen+1)))
			{
				fprintf(stderr,"Couldn't parse state #%d\n",i);
				return NULL;
			}
		}
	}

	pad(padlen); printf("}\n");

	return ptr;
}

static uint8_t *parseTechnique(Technique *technique,uint8_t *ptr,uint8_t *constData,Parameter **objects,int padlen)
{
	uint32_t offset=r32(ptr); ptr+=4;
	technique->name=parseName(constData+offset);

	pad(padlen); printf("Technique %s\n",technique->name);
	pad(padlen); printf("{\n");
	technique->numAnnotations=r32(ptr); ptr+=4;
	technique->numPasses=r32(ptr); ptr+=4;

	uint32_t i;
	if (technique->numAnnotations)
	{
		technique->annotations=malloc(sizeof(Parameter)*technique->numAnnotations);
		for (i=0;i<technique->numAnnotations;i++)
		{
			if (!(ptr=parseAnnotation(&technique->annotations[i],ptr,constData,objects,padlen+1)))
			{
				fprintf(stderr,"Failed to parse annotation #%d\n",i);
				return NULL;
			}
		}
	}
	if (technique->numPasses)
	{
		technique->passes=malloc(sizeof(Pass)*technique->numPasses);
		for (i=0;i<technique->numPasses;i++)
		{
			if (!(ptr=parsePass(&technique->passes[i],ptr,constData,objects,padlen+1)))
			{
				fprintf(stderr,"Failed to parse pass #%d\n",i);
				return NULL;
			}
		}
	}

	pad(padlen); printf("}\n");
	return ptr;
}

static uint8_t *parseData(Parameter *param,uint8_t *ptr,int padlen)
{
	uint32_t size=r32(ptr); ptr+=4;

	switch (param->type)
	{
	case STRING:
		param->data=parseName(ptr-4);
		break;
	case VERTEXSHADER:
		parseVertexShader(ptr,param->data,padlen);
		break;
	case PIXELSHADER:
		parsePixelShader(ptr,param->data,padlen);
		break;
	default:
		fprintf(stderr,"Unhandled type %d\n",param->type);
		return NULL;
	}

	return ptr+((size+3)&~3);
}
static uint8_t *copyData(uint8_t **str,uint8_t *ptr)
{
	uint32_t size=r32(ptr); ptr+=4;
	*str=malloc(size);
	memcpy(*str,ptr,size);
	return ptr+((size+3)&~3);
}

static uint8_t *parseResource(uint8_t *ptr,uint8_t *constData,Parameter *params,Technique *techniques,int padlen)
{
	uint32_t techniqueid=r32(ptr); ptr+=4;
	uint32_t id=r32(ptr); ptr+=4;
	uint32_t elementid=r32(ptr); ptr+=4;
	uint32_t stateid=r32(ptr); ptr+=4;
	uint32_t usage=r32(ptr); ptr+=4;

	Parameter *param=NULL;
	State *state=NULL;


	if (techniqueid==0xffffffff)
	{
		param=&params[id];
		pad(padlen); printf("Resource for Param %s\n",param->name);
		pad(padlen); printf("{\n");
		if (elementid!=0xffffffff && param->numElements)
			param=&param->children[elementid];
		Sampler *sampler=param->data;
		state=&sampler->states[stateid];
	}
	else
	{
		Technique *technique=&techniques[techniqueid];
		Pass *pass=&technique->passes[id];
		pad(padlen); printf("Resource for Technique %s Pass %s\n",technique->name,pass->name);
		pad(padlen); printf("{\n");
		state=&pass->states[stateid];
	}


	param=state->param;
	switch (usage)
	{
	case 0:
		switch (param->type)
		{
		case VERTEXSHADER:
		case PIXELSHADER:
			state->type=CONSTANT;
			if (!(ptr=parseData(param,ptr,padlen+1)))
			{
				fprintf(stderr,"Couldn't parse data\n");
				return NULL;
			}
			break;
		case BOOL:
		case INT:
		case FLOAT:
		case STRING:
			state->type=FXLC;
			if (!(ptr=copyData(param->data,ptr)))
			{
				fprintf(stderr,"Couldn't copy data\n");
				return NULL;
			}
			break;
		default:
			fprintf(stderr,"Unknown type: %d\n",param->type);
			return NULL;
		}
		break;
	case 1:
		state->type=PARAMETER;
		if (!(ptr=copyData(param->data,ptr)))
		{
			fprintf(stderr,"Couldn't copy data\n");
			return NULL;
		}
		break;
	default:
		fprintf(stderr,"Unknown usage: %d\n",usage);
		return NULL;
	}
	pad(padlen); printf("}\n");
	return ptr;
}

static void parseEffect(uint8_t *data,uint32_t len)
{
	if (r32(data)!=0xbcf00bcf)
	{
		fprintf(stderr,"Not a valid effect\n");
		return;
	}
	uint32_t headerLen=r32(data+4);
	data+=headerLen;
	len-=headerLen;

	if (r32(data)!=0xfeff0901)
	{
		fprintf(stderr,"Not a DX9.1 Effect\n");
		return;
	}
	uint32_t constDataLen=r32(data+4);
	uint8_t *constData=data+8;

	uint8_t *ptr=constData+constDataLen;


	uint32_t numParams=r32(ptr); ptr+=4;
	uint32_t numTechniques=r32(ptr); ptr+=4;
	ptr+=4; //unknown
	uint32_t numObjects=r32(ptr); ptr+=4;

	Parameter **objects=malloc(sizeof(Parameter *)*numObjects);

	Parameter *params=malloc(sizeof(Parameter)*numParams);


	printf("Effect\n{\n");

	uint32_t i;
	for (i=0;i<numParams;i++)
	{
		if (!(ptr=parseParameter(&params[i],ptr,constData,objects,1)))
		{
			fprintf(stderr,"Couldn't parse parameter #%d\n",i);
			return;
		}
	}

	Technique *techniques=malloc(sizeof(Technique)*numTechniques);
	for (i=0;i<numTechniques;i++)
	{
		if (!(ptr=parseTechnique(&techniques[i],ptr,constData,objects,1)))
		{
			fprintf(stderr,"Couldn't parse technique #%d\n",i);
			return;
		}
	}

	uint32_t numStrings=r32(ptr); ptr+=4;
	uint32_t numResources=r32(ptr); ptr+=4;

	for (i=0;i<numStrings;i++)
	{
		uint32_t id=r32(ptr); ptr+=4;
		if (!(ptr=parseData(objects[id],ptr,1)))
		{
			fprintf(stderr,"Couldn't parse data #%d\n",i);
			return;
		}
	}

	for (i=0;i<numResources;i++)
	{
		if (!(ptr=parseResource(ptr,constData,params,techniques,1)))
		{
			fprintf(stderr,"Couldn't parse resource #%d\n",i);
			return;
		}
	}
	printf("}\n");
}

int main(int argc,char *argv[])
{
		if (argc!=2)
		{
			fprintf(stderr,"Usage: %s <shader.xnb>\n",argv[0]);
			return -1;
		}
		FILE *f=fopen(argv[1],"rb");
		fseek(f,0,SEEK_END);
		int len=ftell(f);
		fseek(f,0,SEEK_SET);
		uint8_t *raw=malloc(len);
		fread(raw,1,len,f);
		fclose(f);
	
		uint8_t *p=raw;

		uint32_t header=r32(p); p+=4;
		//header is XNBw XNBx XNBm for windows, unix, mac
		if (header!=0x77424e58 && header!=0x78424e58 && header!=0x6d424e58)
		{
			fprintf(stderr,"%s: Not a valid XNB file\n",argv[1]);
			free(raw);
			return -1;
		}
		uint16_t version=r16(p); p+=2;
		int compressed=version&0x8000;
		version&=0xff; //ignore graphics profile
		if (version!=5 && version!=4)
		{
			fprintf(stderr,"%s: Invalid XNB Version\n",argv[1]);
			free(raw);
			return -1;
		}
		uint32_t length=r32(p); p+=4; //length of entire file
		uint8_t *endp=raw+length;
		if (compressed)
		{
			uint32_t decompLength=r32(p); p+=4;
			uint8_t *decomp=malloc(decompLength);
			uint8_t *dp=decomp;
			struct LZXstate *lzx=LZXinit(16);
			while (p<endp)
			{
				uint8_t hi=*p++;
				uint8_t lo=*p++;
				uint16_t compLen=(hi<<8)|lo;
				uint16_t decompLen=0x8000;
				if (hi==0xff)
				{
					hi=lo;
					lo=*p++;
					decompLen=(hi<<8)|lo;
					hi=*p++;
					lo=*p++;
					compLen=(hi<<8)|lo;
				}
				if (compLen==0 || decompLen==0) //done
					break;
				LZXdecompress(lzx,p,dp,compLen,decompLen);
				p+=compLen;
				dp+=decompLen;
			}
			LZXteardown(lzx);
			free(raw);
			raw=decomp;
			p=raw;
			endp=raw+decompLength;
		}

		int numReaders=*p++;
		int i;
		for (i=0;i<numReaders;i++)
		{
			int nameLen=*p++;
			nameLen++;
			p+=nameLen;
			p+=4; //skip reader version
		}
		p+=2; //num shared and res id

		uint32_t size=r32(p); p+=4;
		parseEffect(p,size);

		free(raw);
		return 0;
}
